from logzero import logger as log

import os
import glob
import ROOT
import math
import numpy
import pprint
import utils_noroot      as utnr
import zutils.utils      as zut
import matplotlib.pyplot as plt

from zutils.plot   import plot     as zfp
from fitter        import zfitter

#--------------------------------------
class normalizer:
    '''
    Class used to find normalizations for combinatorial and PRec components
    in the high-q2 signal model
    '''
    #--------------------------------------
    def __init__(self, dset=None, trig=None, model=None, d_val=None, d_var=None):
        self._dset    = dset
        self._trig    = trig
        self._model   = model 
        self._d_val   = d_val
        self._d_var   = d_var

        self._l_flt_par= ['lm_cb', 'mu_cb', 'ncb', 'npr_ee']
        self._dat_path = f'{os.environ["CASDIR"]}/tools/apply_selection/blind_fits/data/v10.21p2'

        self._d_const  = {}
        self._nbins    = 60
        self._hig_mm   = 5360, 5600
        self._low_mm   = 5180, 5200 
        self._hig_ee   = 5450, 5950 
        self._low_ee   = 4450, 5000 

        self._d_pre       = None
        self._out_dir     = None
        self._d_bdt_wp    = None
        self._initialized = False 
    #--------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        s_par       = self._model.get_params()
        self._d_pre = {par.name : par.value().numpy() for par in s_par}

        os.makedirs(f'{self._out_dir}/pdf', exist_ok=True)
        self._d_const= self._prepare_model(self._model)

        self._initialized = True 
    #--------------------------------------
    @property
    def bdt_wp(self):
        return self._d_bdt_wp

    @bdt_wp.setter
    def bdt_wp(self, value):
        '''
        If used, will use a new WP for the sideband data 

        Parameters:
        value (dict): Dictionary mapping BDT to new WP, e.g. {'BDT_cmb' : 0.1, 'BDT_prc' : 0.6}
        '''
        self._d_bdt_wp = value
    #--------------------------------------
    def _get_data_paths(self):
        if   self._dset == 'r1':
            l_dset = [f'{self._dat_path}/2011_{self._trig}', f'{self._dat_path}/2012_{self._trig}']
        elif self._dset == 'r2p1':
            l_dset = [f'{self._dat_path}/2015_{self._trig}', f'{self._dat_path}/2016_{self._trig}']
        elif self._dset == 'all':
            l_dset = [f'{self._dat_path}/{year}_{self._trig}' for year in [2011, 2012, 2015, 2016, 2017, 2018] ]
        else:
            l_dset = [f'{self._dat_path}/{self._dset}_{self._trig}']

        l_wc = [f'{dset}/*.root' for dset in l_dset]

        l_l_path = [ glob.glob(wc) for wc in l_wc ]
        for l_path, wc in zip(l_l_path, l_wc):
            if len(l_path) == 0:
                log.error(f'No file found in {wc}')
                raise

        return l_wc
    #-------------------------------------
    def _filter(self, rdf):
        if self._d_bdt_wp is None:
            log.info('Using nominal BDT WP')
            return rdf
        else:
            log.info('Updating BDT WP')
            pprint.pprint(self._d_bdt_wp)

        cmb_wp = self._d_bdt_wp['BDT_cmb']
        prc_wp = self._d_bdt_wp['BDT_prc']

        rdf = rdf.Filter(f'BDT_cmb > {cmb_wp}', 'CMB')
        rdf = rdf.Filter(f'BDT_prc > {prc_wp}', 'PRC')

        rep = rdf.Report()
        rep.Print()

        return rdf
    #-------------------------------------
    def _get_mass(self, l_dpath):
        rdf = ROOT.RDataFrame(self._trig, l_dpath)
        rdf = self._filter(rdf)

        arr_mass = rdf.AsNumpy(['B_M'])['B_M']

        if self._trig != 'MTOS':
            low, hig    = self._low_ee[1], self._hig_ee[0]
            arr_mass_lo = arr_mass[(arr_mass < low)]
            arr_mass_hi = arr_mass[(arr_mass > hig)]

            arr_mass_cn = arr_mass[(arr_mass > low) & (arr_mass < hig)]
            arr_mass_cn = numpy.random.uniform(low=low, high=hig, size=arr_mass_cn.shape)

            arr_mass = numpy.concatenate([arr_mass_lo, arr_mass_cn, arr_mass_hi])
        else:
            log.warning(f'Unblinding: {self._trig}')

        return arr_mass
    #--------------------------------------
    def _get_data(self):
        dat_path = f'{self._out_dir}/data/{self._dset}_{self._trig}.json'
        if os.path.isfile(dat_path):
            log.info(f'Loading cached data from: {dat_path}')
            l_dat = utnr.load_json(dat_path)
            return numpy.array(l_dat)

        os.makedirs(f'{self._out_dir}/data', exist_ok=True)

        l_dpath = self._get_data_paths()
        arr_dat = self._get_mass(l_dpath)

        log.info(f'Saving to: {dat_path}')
        utnr.dump_json(arr_dat.tolist(), dat_path) 
        self._plot_data(arr_dat)

        return arr_dat
    #--------------------------------------
    def _plot_data(self, arr_mass):
        if not self._out_dir:
            return

        rng_ee = 4800, 6000
        rng_mm = 5100, 5600
        rng = rng_mm if self._trig == 'MTOS' else rng_ee
        plt.hist(arr_mass, range=rng, bins=50, histtype='step')
        plot_path = f'{self._out_dir}/data/{self._dset}_{self._trig}.png'
        log.info(f'Saving to: {plot_path}')
        plt.savefig(plot_path)
        plt.close('all')
    #--------------------------------------
    def _prepare_model(self, pdf):
        s_par = pdf.get_params()
        for par in s_par:
            par.floating = False

        d_const = {}
        for par in s_par:
            for flt_par in self._l_flt_par:
                if par.name.startswith(flt_par):
                    par.floating = True
                else:
                    continue

                if par.name in self._d_val:
                    var = self._d_var[par.name]
                    val = self._d_val[par.name]
                    d_const[par.name] = val, math.sqrt(var)

        return d_const
    #--------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot create directory: {value}')
            raise

        self._out_dir = value
    #--------------------------------------
    def _get_stats(self, arr_dat):
        if self._trig != 'MTOS':
            return ''

        arr_flg = (arr_dat > 5180) & (arr_dat < 5400)
        arr_dat = arr_dat[arr_flg]
        ndata   = float(arr_dat.size)

        s_par     = self._model.get_params(floating=False)
        [pst_val] = [ par.value().numpy() for par      in s_par               if par.name.startswith('nsg_mm_') ]
        [pre_val] = [ val                 for nam, val in self._d_pre.items() if nam.startswith('nsg_mm_')      ]

        v1 = f'Data: {ndata:.0f} $m\in [5180, 5400]$'
        v2 = f'Fitted: {pst_val:.0f}'
        v3 = f'Expected: {pre_val:.0f}'

        return f'{v1}\n{v2}\n{v3}'
    #--------------------------------------
    def _get_pdf_names(self):
        d_leg         = {}
        d_leg['prc']  = r'$\psi(2S) + c\bar{c}$'
        d_leg['bpks'] = r'$B^+\to K^{*+}e^+e^-$'
        d_leg['bdks'] = r'$B^0\to K^{*0}e^+e^-$'
        d_leg['bsph'] = r'$B_s\to \phi e^+e^-$'
        d_leg['bpk1'] = r'$B^+\to K_{1}e^+e^-$'
        d_leg['bpk2'] = r'$B^+\to K_{2}e^+e^-$'

        return d_leg
    #--------------------------------------
    def _plot_fit(self, data=None, name=None, result=None, l_range=None, stacked=None):
        if self._trig == 'MTOS':
            l_range = None
            log.warning(f'Plotting unblinded data for {self._trig}')

        stats = self._get_stats(data)

        obj= zfp(data=data, model=self._model, result=result)
        obj.plot(skip_pulls=False, nbins=self._nbins, ranges=l_range, d_leg=self._get_pdf_names(), stacked=stacked, ext_text=stats)
        obj.axs[0].grid()

        if   '_all_' in name and 'MTOS' in name:
            obj.axs[0].set_ylim(0,  250)
        elif '_all_' in name and 'ETOS' in name:
            obj.axs[0].set_ylim(0,  120)
        elif '_all_' in name and 'GTIS' in name:
            obj.axs[0].set_ylim(0,   50)

        obj.axs[1].set_ylim(-5, 5)
        obj.axs[1].axhline(0, color='r')

        os.makedirs(f'{self._out_dir}/fits', exist_ok=True)
        plot_path = f'{self._out_dir}/fits/{name}.png'
        log.info(f'Saving to: {plot_path}')
        plt.legend(loc='upper right')
        plt.savefig(plot_path)
        plt.close('all')
    #--------------------------------------
    def get_fit_result(self):
        self._initialize()

        arr_mass= self._get_data()

        low   = self._low_mm if self._trig == 'MTOS' else self._low_ee
        hig   = self._hig_mm if self._trig == 'MTOS' else self._hig_ee

        zut.print_pdf(self._model, txt_path=f'{self._out_dir}/pdf/pre_{self._dset}_{self._trig}.txt', d_const=self._d_const)
        obj   = zfitter(self._model, arr_mass)
        res   = obj.fit(ranges=[low, hig], d_const=self._d_const)
        zut.print_pdf(self._model, txt_path=f'{self._out_dir}/pdf/pos_{self._dset}_{self._trig}.txt', d_const=self._d_const)

        self._plot_fit(data=arr_mass, name=f'{self._trig}_{self._dset}_stk', result=res, l_range=[low, hig], stacked= True)
        self._plot_fit(data=arr_mass, name=f'{self._trig}_{self._dset}_ovr', result=res, l_range=[low, hig], stacked=False)

        res.hesse()
        res.freeze()
        print(res)

        return res
#--------------------------------------




from logzero     import logger as log
from zutils.plot import plot   as zfp

import matplotlib.pyplot as plt
import zutils.utils      as zut
import utils_noroot      as utnr
import pandas            as pnd

import os
import tqdm
import zfit
import numpy
import time

#---------------------------------
class analyzer:
    '''
    This tool is meant to provide diagnostic information on an extended PDF as
    implemented by Zfit
    '''
    #---------------------------------
    def __init__(self, pdf=None, d_const=None, nev_fac=1):
        self._pdf     = pdf
        self._d_cns   = d_const
        self._nev_fac = nev_fac

        self._dat     = None
        self._nll     = None
        self._mnz     = None
        self._out_dir = None
        self._min_obs = None
        self._max_obs = None
        self._l_nam   = None
        self._d_inival= dict()
        self._d_info  = dict()

        self._initialized = False
    #---------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        if value is None:
            return

        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot make: {value}')
            raise

        self._out_dir = value
    #---------------------------------
    def _initialize(self):
        if self._initialized:
            return

        zfit.settings.changed_warnings.all = False

        self._check_pdf()
        log.debug('Creating sampler')
        self._dat = self._pdf.create_sampler(fixed_params=False)
        self._prepare_fit()

        self._initialized = True
    #---------------------------------
    def _get_const(self):
        if self._d_cns is None:
            return

        l_par = [ par                      for par in self._pdf.get_params() if par.name in self._d_cns] 

        if len(l_par) == 0:
            log.warning(f'Found no floating parameters to constrain')
            return

        l_val = [ self._d_cns[par.name][0] for par in l_par ] 
        l_err = [ self._d_cns[par.name][1] for par in l_par ] 

        return zfit.constraint.GaussianConstraint(l_par, l_val, l_err)
    #---------------------------------
    def _prepare_fit(self):
        log.debug('Preparing fit')
        cons      = self._get_const()
        self._nll = zfit.loss.ExtendedUnbinnedNLL(model=self._pdf, data=self._dat, constraints=cons)
        self._mnz = zfit.minimize.Minuit()
    #---------------------------------
    def _check_pdf(self):
        log.debug('Checking PDF')
        if self._pdf is None:
            log.error(f'Missing PDF')
            raise

        if not self._pdf.is_extended:
            log.error(f'Only extended PDFs supported')
            raise

        s_par       = self._pdf.get_params()
        self._l_nam = [ par.name for par in s_par ]

        self._d_inival = { par.name : par.value().numpy() for par      in s_par }
        d_yield        = { key      : self._nev_fac * val for key, val in self._d_inival.items() if key.startswith('n')}

        self._d_inival.update(d_yield)
    #---------------------------------
    def _reset_pdf(self):
        s_par = self._pdf.get_params()
        for par in s_par:
            ini_val = self._d_inival[par.name]
            par.set_value(ini_val)
    #---------------------------------
    def _plot_data(self, i_fit, nfit):
        if self._out_dir is None:
            return
        elif i_fit == 10 or (i_fit == nfit - 1):
            plot_path = f'{self._out_dir}/data.png'
            plt.savefig(plot_path)
            plt.close('all')
            return
        elif i_fit > 10:
            return

        arr_val = self._dat.numpy().flatten()
        if (self._min_obs is None) or (self._max_obs is None):
            self._min_obs = min(arr_val)
            self._max_obs = max(arr_val)

        plt.hist(arr_val, bins=30, range=[self._min_obs, self._max_obs], histtype='step')
    #---------------------------------
    def _plot_fit(self, i_fit):
        if self._out_dir is None:
            return
        elif i_fit > 10:
            return

        obj=zfp(data=self._dat, model=self._pdf, result=None)
        obj.plot(nbins=50, stacked=True)

        plot_path = f'{self._out_dir}/fits/{i_fit:03}.png'
        utnr.make_path_dir(plot_path)

        plt.savefig(plot_path)
        plt.close('all')
    #---------------------------------
    def _run_minimization(self, nfit, kind=None):
        l_res = []

        log.info(f'Running fits, for {kind}')
        for i_fit in tqdm.trange(nfit, ascii=' -'):
            self._reset_pdf()
            zut.print_pdf(self._pdf, txt_path=f'{self._out_dir}/model/pre_{i_fit:03}.txt', d_const=self._d_cns)
            self._dat.resample()
            #self._plot_fit(i_fit)
            res=self._mnz.minimize(self._nll)
            zut.print_pdf(self._pdf, txt_path=f'{self._out_dir}/model/pos_{i_fit:03}.txt', d_const=self._d_cns)
            l_res.append(res)

        return l_res
    #---------------------------------
    def _run_hesse(self, l_res):
        log.info('Running Hesse')
        for res in tqdm.tqdm(l_res, ascii=' -'):
            try:
                res.hesse()
            except:
                log.warning('Hesse failed, skipping')
                continue
    #---------------------------------
    def speed(self, nfit=100):
        '''
        Run fit multiple times and time it

        Parameters:
        -------------------
        nfit (int): Number of fits over which to average the fitting time
        '''
        self._initialize()
        log.info(f'Using {nfit} fits')


        t_1   = time.time()
        l_res = self._run_minimization(nfit, kind='speed')
        t_2   = time.time()
        self._run_hesse(l_res)
        t_3   = time.time()

        t_fit = (t_2 - t_1) / nfit
        t_hes = (t_3 - t_2) / nfit

        self._d_info['Fit/second'] = t_fit
        self._d_info['Hes/second'] = t_hes 
        self._d_info['#Fits']      = nfit 

        log.info(f'Fit takes: {t_fit:.3} seconds')
        log.info(f'Hesse takes: {t_hes:.3} seconds')

        self._finalize()
    #---------------------------------
    def fit(self, nfit=100):
        '''
        Used to make toy fits and make pull distributions 

        Parameters
        ------------------
        nfit (int): Number of fits over which to average the fitting time
        '''
        self._initialize()

        l_res = self._run_minimization(nfit, kind='pulls')
        self._run_hesse(l_res)

        df_ini, df_val, df_err = self._get_fit_df(l_res)

        return df_ini, df_val, df_err
    #---------------------------------
    def _get_val_err(self, res):
        l_ini = []
        l_val = []
        l_err = []

        res.freeze()
        for nam in self._l_nam:
            d_val = res.params[nam]
            ini   = self._d_inival[nam]
            val   = d_val['value']
            err   = d_val['hesse']['error'] 

            l_ini.append(ini)
            l_val.append(val)
            l_err.append(err)

        return l_ini, l_val, l_err
    #---------------------------------
    def _get_fit_df(self, l_res):
        df_ini = pnd.DataFrame(columns=self._l_nam)
        df_val = pnd.DataFrame(columns=self._l_nam)
        df_err = pnd.DataFrame(columns=self._l_nam)

        for res in l_res:
            l_ini, l_val, l_err = self._get_val_err(res)

            df_ini = utnr.add_row_to_df(df_ini, l_ini)
            df_val = utnr.add_row_to_df(df_val, l_val)
            df_err = utnr.add_row_to_df(df_err, l_err)

        return df_ini, df_val, df_err
    #---------------------------------
    def sampling_speed(self, nsample=100):
        #self._initialize()

        [[min_x]], [[max_x]] = self._pdf.space.limits
        arr_xval = numpy.linspace(min_x, max_x, nsample)

        log.info(f'Sampling {self._pdf}')
        [ self._pdf.pdf(xval) for xval in tqdm.tqdm(arr_xval, ascii=' -') ]

        for pdf in self._pdf.pdfs:
            log.info(f'Sampling {pdf.name}')
            [ pdf.pdf(xval) for xval in tqdm.tqdm(arr_xval, ascii=' -') ]
    #---------------------------------
    def _finalize(self):
        if self._out_dir is None:
            return

        out_path = f'{self._out_dir}/info.json'
        utnr.dump_json(self._d_info, out_path)

        log.info(f'Saving to: {out_path}')
#---------------------------------

